# -*- coding: utf-8 -*-
# Wendelin.bigfile | WCFS part of BigFile ZODB backend
# Copyright (C) 2014-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# cython: language_level=2
# distutils: language=c++

"""XXX"""

cdef extern from "wcfs/internal/wcfs.h":
    pass

cdef extern from "bigfile/_bigfile.h":
    struct PyBigFile:
        pass
    ctypedef extern class wendelin.bigfile._bigfile.BigFile[object PyBigFile]:
        pass

import wcfs as pywcfs
from wcfs.internal cimport _wcfs as wcfs
from golang cimport nil

from ZODB.Connection import Connection as ZConnection
from ZODB.utils import u64
from wendelin.lib.zodb import zconn_at


# _ZBigFile is base class for ZBigFile that provides BigFile-line base.
#
# The other base line is from Persistent. It is not possible to inherit from
# both Persistent and BigFile at the same time since both types are C types and
# their layouts conflict.
#
# _ZBigFile:
#
# - redirects loadblk/storeblk calls to ZBigFile.
# - provides blkmmapper with WCFS integration.
cdef public class _ZBigFile(BigFile) [object _ZBigFile, type _ZBigFile_Type]:
    cdef object     zself   # reference to ZBigFile
    cdef wcfs.FileH wfileh  # WCFS file handle. Initially nil, opened by blkmmapper

    # _new creates new _ZBigFile associated with ZBigFile zself.
    # XXX Cython does not allow __new__ nor to change arguments passed to __cinit__ / __init__
    @staticmethod
    def _new(zself, blksize):
        cdef _ZBigFile obj = _ZBigFile.__new__(_ZBigFile, blksize)
        obj.zself  = zself
        obj.wfileh = nil
        return obj

    def __dealloc__(_ZBigFile zf):
        zf.wfileh = nil


    # redirect load/store to main class
    def loadblk(self, blk, buf):    return self.zself.loadblk(blk, buf)
    def storeblk(self, blk, buf):   return self.zself.storeblk(blk, buf)

    # blkmmapper returns pycapsule with virtmem mmapper for the file.
    #
    # XXX
    #
    # it is called from PyBigFile.fileh_open(mmap_overlay=True) and .zfile
    # should be already associated with jar and have oid.
    def blkmmapper(_ZBigFile zf):
        cdef wcfs.PyFileH pywfileh
        if zf.wfileh == nil:
            zconn = zf.zself._p_jar
            assert zconn is not None
            # XXX locking? or rely on that ZODB objects for must be used from under 1 thread only?

            # join zconn to wconn; link to wconn from _ZBigFile
            pywconn   = pywconnOf(zconn)
            pywfileh  = pywconn.open(zf.zself._p_oid)
            zf.wfileh = pywfileh.wfileh

        # return XXX


# functions that we give to virtmem bigfile_ops .mmap*
cdef extern from * nogil:
    pass



# pywconnOf establishes and returns (py) wcfs.Conn associated with zconn.
# XXX wconn is maintained and kept in sync with zconn.
# XXX simultaneous call?
# XXX move -> .py ?
cdef wcfs.PyConn pywconnOf(zconn):
    assert isinstance(zconn, ZConnection)
    assert zconn.opened     # XXX needed

    # XXX locking
    wconn = getattr(zconn, '_wcfs_wconn', None)
    if wconn is not None:
        return wconn

    # zconn is not yet associated with wconn
    zstor = zconn.db().storage
    zurl  = pywcfs.zstor_2zurl(zstor)
    wc    = pywcfs.join(zurl)
    wconn = wc.connect(zconn_at(zconn))

    # XXX register zconn synchronizer to wconn.resync on zconn open
    # XXX zconn goes away -> wconn close & free
    zconn._wcfs_wconn = wconn
    return wconn
